""" Provide a class with a context manager to help avoid overloading web servers. """

import asyncio
import logging
import os
import random
import sqlite3
import time
import urllib.parse


class AccessRateWatcher:

    """Access rate limiter, supporting concurrent access by threads and/or processes."""

    def __init__(
        self, db_filepath, url, min_delay_between_accesses, *, jitter_range_ms=None, logger=logging.getLogger()
    ):
        self.domain = urllib.parse.urlsplit(url).netloc
        self.min_delay_between_accesses = min_delay_between_accesses
        self.jitter_range_ms = jitter_range_ms
        self.logger = logger
        os.makedirs(os.path.dirname(db_filepath), exist_ok=True)
        self.connection = sqlite3.connect(db_filepath)
        with self.connection:
            self.connection.executescript(
                """CREATE TABLE IF NOT EXISTS access_timestamp (domain TEXT PRIMARY KEY,
                                                                                    timestamp FLOAT NOT NULL);"""
            )
        self.lock = None

    async def waitAccessAsync(self):
        """Wait the needed time before sending a request to honor rate limit."""
        if self.lock is None:
            self.lock = asyncio.Lock()

        async with self.lock:
            while True:
                last_access_ts = self.__getLastAccess()
                if last_access_ts is not None:
                    now = time.time()
                    last_access_ts = last_access_ts[0]
                    time_since_last_access = now - last_access_ts
                    if time_since_last_access < self.min_delay_between_accesses:
                        time_to_wait = self.min_delay_between_accesses - time_since_last_access
                        if self.jitter_range_ms is not None:
                            time_to_wait += random.randint(*self.jitter_range_ms) / 1000
                        self.logger.debug(
                            "Sleeping for %.2fms because of rate limit for domain %s"
                            % (time_to_wait * 1000, self.domain)
                        )
                        await asyncio.sleep(time_to_wait)

                access_time = time.time()
                self.__access(access_time)

                # now we should be good... except if another process did the same query at the same time
                # the database serves as an atomic lock, query again to be sure the last row is the one
                # we just inserted
                last_access_ts = self.__getLastAccess()
                if last_access_ts[0] == access_time:
                    break

    def __getLastAccess(self):
        with self.connection:
            return self.connection.execute(
                """SELECT timestamp
                                        FROM access_timestamp
                                        WHERE domain = ?;""",
                (self.domain,),
            ).fetchone()

    def __access(self, ts):
        """Record an API access."""
        with self.connection:
            self.connection.execute(
                "INSERT OR REPLACE INTO access_timestamp (timestamp, domain) VALUES (?, ?)", (ts, self.domain)
            )



""" Itunes cover source. """

import collections
import json

from sacad.cover import SUPPORTED_IMG_FORMATS as EXTENSION_FORMAT
from sacad.cover import CoverImageFormat, CoverImageMetadata, CoverSourceQuality, CoverSourceResult
from sacad.sources.base import CoverSource


class ItunesCoverSourceResult(CoverSourceResult):

    """Itunes search cover result."""

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            source_quality=CoverSourceQuality.NO_UNRELATED_RESULT_RISK | CoverSourceQuality.EXACT_SEARCH,
            **kwargs,
        )


class ItunesCoverSource(CoverSource):

    """Itunes cover source."""

    SEARCH_URL = "https://itunes.apple.com/search"

    def __init__(self, *args, **kwargs):
        # https://stackoverflow.com/questions/12596300/itunes-search-api-rate-limit
        super().__init__(*args, min_delay_between_accesses=3, **kwargs)

    def getSearchUrl(self, album, artist):
        """See CoverSource.getSearchUrl."""
        url_params = collections.OrderedDict()
        url_params["media"] = "music"
        url_params["entity"] = "album"
        url_params["term"] = f"{artist} {album}"
        return __class__.assembleUrl(__class__.SEARCH_URL, url_params)

    async def parseResults(self, api_data):
        """See CoverSource.parseResults."""
        json_data = json.loads(api_data)

        results = []
        for rank, result in enumerate(json_data["results"], 1):
            thumbnail_url = result["artworkUrl60"]
            base_img_url = result["artworkUrl60"].rsplit("/", 1)[0]
            url_found = False
            for img_size in (5000, 1200, 600):
                for img_format in (CoverImageFormat.PNG, CoverImageFormat.JPEG):
                    suffix = "-100.jpg" if (img_format is CoverImageFormat.JPEG) else ".png"
                    img_url = f"{base_img_url}/{img_size}x{img_size}{suffix}"
                    if await self.probeUrl(img_url):
                        url_found = True
                        break
                if url_found:
                    break
            else:
                img_url = result["artworkUrl100"]
                img_format = EXTENSION_FORMAT[img_url.rsplit(".", 1)[-1]]
                img_size = 100
            result = ItunesCoverSourceResult(
                img_url,
                (img_size, img_size),
                img_format,
                thumbnail_url=thumbnail_url,
                source=self,
                rank=rank,
                check_metadata=CoverImageMetadata.NONE,
            )
            results.append(result)

        return results



""" LastFM cover source. """

import collections
import os.path
import string
import xml.etree.ElementTree

from sacad.cover import SUPPORTED_IMG_FORMATS, CoverImageMetadata, CoverSourceQuality, CoverSourceResult
from sacad.sources.base import MAX_THUMBNAIL_SIZE, CoverSource


class LastFmCoverSourceResult(CoverSourceResult):

    """LastFM cover search result."""

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            source_quality=CoverSourceQuality.EXACT_SEARCH | CoverSourceQuality.NO_UNRELATED_RESULT_RISK,
            **kwargs,
        )


class LastFmCoverSource(CoverSource):

    """
    Cover source using the official LastFM API.

    http://www.lastfm.fr/api/show?service=290
    """

    BASE_URL = "https://ws.audioscrobbler.com/2.0/"
    API_KEY = "2410a53db5c7490d0f50c100a020f359"

    SIZES = {
        "small": (34, 34),
        "medium": (64, 64),
        "large": (174, 174),
        "extralarge": (300, 300),
        "mega": (600, 600),
    }  # this is actually between 600 and 900, sometimes even more (ie 1200)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, min_delay_between_accesses=0.1, **kwargs)

    def getSearchUrl(self, album, artist):
        """See CoverSource.getSearchUrl."""
        # build request url
        params = collections.OrderedDict()
        params["method"] = "album.getinfo"
        params["api_key"] = __class__.API_KEY
        params["album"] = album
        params["artist"] = artist

        return __class__.assembleUrl(__class__.BASE_URL, params)

    def processQueryString(self, s):
        """See CoverSource.processQueryString."""
        char_blacklist = set(string.punctuation)
        char_blacklist.remove("'")
        char_blacklist.remove("&")
        char_blacklist = frozenset(char_blacklist)
        return __class__.unpunctuate(s.lower(), char_blacklist=char_blacklist)

    async def parseResults(self, api_data):
        """See CoverSource.parseResults."""
        results = []

        # get xml results list
        xml_text = api_data.decode("utf-8")
        xml_root = xml.etree.ElementTree.fromstring(xml_text)
        status = xml_root.get("status")
        if status != "ok":
            raise Exception(f"Unexpected Last.fm response status: {status}")
        img_elements = xml_root.findall("album/image")

        # build results from xml
        thumbnail_url = None
        thumbnail_size = None
        for img_element in img_elements:
            img_url = img_element.text
            if not img_url:
                # last.fm returns empty image tag for size it does not have
                continue
            lfm_size = img_element.get("size")
            if lfm_size == "mega":
                check_metadata = CoverImageMetadata.SIZE
            else:
                check_metadata = CoverImageMetadata.NONE
            try:
                size = __class__.SIZES[lfm_size]
            except KeyError:
                continue
            if (size[0] <= MAX_THUMBNAIL_SIZE) and ((thumbnail_size is None) or (size[0] < thumbnail_size)):
                thumbnail_url = img_url
                thumbnail_size = size[0]
            format = os.path.splitext(img_url)[1][1:].lower()
            format = SUPPORTED_IMG_FORMATS[format]
            results.append(
                LastFmCoverSourceResult(
                    img_url, size, format, thumbnail_url=thumbnail_url, source=self, check_metadata=check_metadata
                )
            )

        return results
